HudMap = LibStub("AceAddon-3.0"):NewAddon("HudMap", "AceEvent-3.0", "AceHook-3.0")
local mod = HudMap
local L = LibStub("AceLocale-3.0"):GetLocale("HudMap")
local db

BINDING_HEADER_HUDMAP 			= L["HudMap"]
BINDING_NAME_HUDMAP_ZOOMIN 	= L["Zoom In"]
BINDING_NAME_HUDMAP_ZOOMOUT = L["Zoom Out"]
BINDING_NAME_TOGGLE_HUDMAP 	= L["Toggle HudMap"]

--[[ Default upvals 
     This has a slight performance benefit, but upvalling these also makes it easier to spot leaked globals. ]]--
local _G = _G.getfenv(0)
local wipe, type, pairs, tinsert, tremove, tonumber = _G.wipe, _G.type, _G.pairs, _G.tinsert, _G.tremove, _G.tonumber
local math, math_abs, math_pow, math_sqrt, math_sin, math_cos, math_atan2 = _G.math, _G.math.abs, _G.math.pow, _G.math.sqrt, _G.math.sin, _G.math.cos, _G.math.atan2
local error, rawset, rawget, print = _G.error, _G.rawset, _G.rawget, _G.print
local tonumber, tostring = _G.tonumber, _G.tostring
local getmetatable, setmetatable, pairs, ipairs, select, unpack = _G.getmetatable, _G.setmetatable, _G.pairs, _G.ipairs, _G.select, _G.unpack
--[[ -------------- ]]--

local CallbackHandler = LibStub:GetLibrary("CallbackHandler-1.0")
local AceConfig = LibStub("AceConfig-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")
local updateFrame= CreateFrame("Frame")
local updateRotations, zoneScalingData, zoneOverrides, minimapSize, indoors, Point, Edge
local followedUnits = {}

local SN = setmetatable({}, {__index = function(t, k)
	local n = GetSpellInfo(k)
	rawset(t, k, n)
	return n
end})

mod.SN = SN

local GetNumRaidMembers, GetNumPartyMembers = _G.GetNumRaidMembers, _G.GetNumPartyMembers
local GetCVar, GetTime, UIParent = _G.GetCVar, _G.GetTime, _G.UIParent
local UnitExists, UnitIsUnit = _G.UnitExists, _G.UnitIsUnit

local HudMapMinimap, HudMapStandaloneCluster = _G.HudMapMinimap, _G.HudMapStandaloneCluster

local textureLookup = {
	diamond 	= [[Interface\TARGETINGFRAME\UI-RAIDTARGETINGICON_3.BLP]],
	star 			= [[Interface\TARGETINGFRAME\UI-RaidTargetingIcon_1.blp]],
	circle  	= [[Interface\TARGETINGFRAME\UI-RaidTargetingIcon_2.blp]],
	triangle 	= [[Interface\TARGETINGFRAME\UI-RaidTargetingIcon_4.blp]],
	moon			= [[Interface\TARGETINGFRAME\UI-RaidTargetingIcon_5.blp]],
	square		= [[Interface\TARGETINGFRAME\UI-RaidTargetingIcon_6.blp]],
	cross			= [[Interface\TARGETINGFRAME\UI-RaidTargetingIcon_7.blp]],
	skull			= [[Interface\TARGETINGFRAME\UI-RaidTargetingIcon_8.blp]],
	cross2		= [[Interface\RAIDFRAME\ReadyCheck-NotReady.blp]],
	check			= [[Interface\RAIDFRAME\ReadyCheck-Ready.blp]],
	question	= [[Interface\RAIDFRAME\ReadyCheck-Waiting.blp]],
	targeting = [[Interface\Minimap\Ping\ping5.blp]],
	highlight = [[Interface\AddOns\HudMap\assets\alert_circle]],
	radius 		= [[Interface\AddOns\HudMap\assets\radius]],
	radius_lg	= [[Interface\AddOns\HudMap\assets\radius_lg]],
	timer			= [[Interface\AddOns\HudMap\assets\timer]],
	glow			= [[Interface\GLUES\MODELS\UI_Tauren\gradientCircle]],
	party     = [[Interface\MINIMAP\PartyRaidBlips]],
	tank      = [[Interface\AddOns\HudMap\assets\roles]],
	dps	      = [[Interface\AddOns\HudMap\assets\roles]],
	healer    = [[Interface\AddOns\HudMap\assets\roles]],
	ring			= [[SPELLS\CIRCLE]],
	-- tank      = [[Interface\LFGFrame\LFGRole_BW]],
	-- dps	      = [[Interface\LFGFrame\LFGRole_BW]],
	-- healer    = [[Interface\LFGFrame\LFGRole_BW]]
}

local texBlending = {
	highlight = "ADD",
	targeting = "ADD",
	glow 			= "ADD",
	ring      = "ADD"
	-- timer			= "ADD",
}


local options = {
	type = "group",
	args = {}
}
mod.options = options

local coreOptions = {
	type = "group",
	name = L["General"],
	args = {
		adaptiveZoom = {
			type = "group",
			name = L["Adaptive Zoom"],
			inline = true,
			disabled = function() return not db.useAdaptiveZoom end,
			args = {
				interestRadius = {
					type = "range",
					name = L["Interest Radius"],
					min = 20,
					max = 200,
					step = 5,
					bigStep = 5,
					get = function()
						return db.interestRadius
					end,
					set = function(info, v)
						db.interestRadius = v
					end,
					order = 101,
				},
				adjust = {
					type = "execute",
					name = L["Set Visible Area"],
					func = function() mod:SetArea() end,
					order = 200
				},
				reset = {
					type = "execute",
					name = L["Reset Area"],
					func = function()
						db.canvasX = nil
						db.canvasY = nil
						mod:UpdateCanvasPosition()
					end,					
					order = 201
				},
			}
		},
		staticZoom = {
			type = "group",
			name = L["Static Zoom"],
			inline = true,
			disabled = function() return db.useAdaptiveZoom end,
			args = {
				zoomLevel = {
					type = "range",
					name = L["Fixed Zoom"],
					min = 20,
					max = 200,
					step = 5,
					bigStep = 5,
					get = function()
						return db.zoomLevel
					end,
					set = function(info, v)
						db.zoomLevel = v
						mod:SetZoom(v)
					end,
					order = 102,					
				},
				zoomInBinding = {
					type = "keybinding",
					name = L["Zoom In"],
					order = 105,
					get = function() return GetBindingKey("HUDMAP_ZOOMIN") end,
					set = function(info, v)
						SetBinding(v, "HUDMAP_ZOOMIN")
						SaveBindings(GetCurrentBindingSet())
					end
				},
				zoomOutBinding = {
					type = "keybinding",
					name = L["Zoom Out"],
					order = 106,
					get = function() return GetBindingKey("HUDMAP_ZOOMOUT") end,
					set = function(info, v)
						SetBinding(v, "HUDMAP_ZOOMOUT")
						SaveBindings(GetCurrentBindingSet())
					end
				},
			}
		},
		general = {
			type = "group",
			name = L["General Options"],
			order = 1,
			inline = true,
			args = {
				useAutozoom = {
					type = "toggle",
					name = L["Use Adaptive Zoom"],
					get = function()
						return db.useAdaptiveZoom
					end,
					set = function(info, v)
						db.useAdaptiveZoom = v
					end,
					order = 100
				},
				alpha = {
					type = "range",
					name = L["Master Opacity"],
					order = 6,
					min = 0,
					max = 1,
					step = 0.01,
					bigStep = 0.01,
					get = function()
						return db.alpha
					end,
					set = function(info, v)
						db.alpha = v
						mod.canvas:SetAlpha(v)
					end,
					order = 101
				},
				binding = {
					type = "keybinding",
					name = L["Toggle Binding"],
					order = 2,
					get = function()
						return GetBindingKey("TOGGLE_HUDMAP")
					end,
					set = function(info, v)
						SetBinding(v, "TOGGLE_HUDMAP")
						SaveBindings(GetCurrentBindingSet())
					end,
					order = 102
				},
			}
		}
	}
}

local defaults = {
	profile = {
		useGatherMate = true,
		useQuestHelper = true,
		useRoutes = true,
		hudColor = {},
		textColor = {r = 0.5, g = 1, b = 0.5, a = 1},
		scale = 6,
		alpha = 1,
		maxSize = UIParent:GetHeight() * 0.48,
		maxSizeSet = false,
		interestRadius = 100,
		zoomLevel = 100,
		enabled = true,
		useAdaptiveZoom = true,
	}
}

local configFrame
do
	configFrame = CreateFrame("Frame", "HudMapConfigFrame", UIParent)
	local closer
	configFrame:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background"})
	configFrame:SetPoint("CENTER")
	configFrame:SetBackdropColor(0, 0.5, 0, 0.5)
	configFrame:Hide()
	configFrame.testDots = {}
	configFrame:SetScript("OnHide", function(self)
		for _, dot in ipairs(self.testDots) do
			dot:Free()
		end
	end)
	configFrame:SetMinResize(100, 100)
	configFrame:SetResizable(true)
	configFrame:SetMovable(true)
	configFrame:EnableMouse(true)
	local onUpdate = function(self)
		local x, y = self:GetCenter()
		local x2, y2 = UIParent:GetCenter()
		db.canvasX = x - x2
		db.canvasY = y - y2
		mod:UpdateCanvasPosition()
	end
	configFrame:SetScript("OnMouseDown", function(self, button)
		if button == "LeftButton" then
			self:StartMoving()
			self:SetScript("OnUpdate", onUpdate)
		end
	end)
	configFrame:SetScript("OnMouseUp", function(self)
		self:StopMovingOrSizing()
		self:SetScript("OnUpdate", nil)
	end)
	configFrame:SetScript("OnSizeChanged", function(self)
		if not self.lastSize or self.noForceResize then
			self.noForceResize = false
			return
		end
		local xd = math.abs(self:GetWidth() - self.lastSize)
		local yd = math.abs(self:GetHeight() - self.lastSize)
		self.noForceResize = true
		local size = xd > yd and self:GetWidth() or self:GetHeight()		
		self.lastSize = size
		self:SetSize(size, size)
		db.maxSize = size / 2
		db.maxSizeSet = true
		self:ClearAllPoints()
		self:SetPoint("CENTER", UIParent, "CENTER", db.canvasX, db.canvasY)
	end)
	
	local locations = {"RIGHT", "BOTTOM"}
	for index, location in ipairs(locations) do
		local arrow = CreateFrame("Button", nil, configFrame)
		arrow:SetNormalTexture([[Interface\GLUES\COMMON\Glue-LeftArrow-Button-Up.blp]])
		arrow:SetHighlightTexture([[Interface\GLUES\COMMON\Glue-LeftArrow-Button-Highlight.blp]])
		arrow:SetPushedTexture([[Interface\GLUES\COMMON\Glue-LeftArrow-Button-Down.blp]])
		local rot = math.pi * (1 - ((index - 1) * 0.5))
		arrow:GetNormalTexture():SetRotation(rot)
		arrow:GetHighlightTexture():SetRotation(rot)
		arrow:GetPushedTexture():SetRotation(rot)
		arrow:SetPoint(location)
		arrow:SetSize(32, 32)
		arrow:SetScript("OnMouseDown", function(self)
			self:GetParent():StartSizing()			
		end)
		arrow:SetScript("OnMouseUp", function(self)
			self:GetParent():StopMovingOrSizing()
		end)
	end
	
	local closer = CreateFrame("Button", nil, configFrame)
	closer:SetNormalTexture([[Interface\BUTTONS\UI-Panel-MinimizeButton-Up.blp]])
	closer:SetHighlightTexture([[Interface\BUTTONS\UI-Panel-MinimizeButton-Highlight.blp]])
	closer:SetPushedTexture([[Interface\BUTTONS\UI-Panel-MinimizeButton-Down.blp]])
	closer:SetPoint("TOPRIGHT")
	closer:SetSize(32, 32)
	closer:SetScript("OnClick", function(self)
		self:GetParent():Hide()
	end)
end

local LDB = LibStub("LibDataBroker-1.1", true)
if LDB then
	local LDB_Plugin = CreateFrame("Frame", "Broker_HudMap")
	LDB_Plugin.obj = LDB:NewDataObject("Broker_HudMap", {type = "data source", icon = [[Interface\Icons\INV_Misc_Map03]], label = "HudMap"})

	function LDB_Plugin.obj.OnClick(self, button)
		-- print(button)
		if button == "RightButton" then
			HudMap:SetArea()
		-- elseif button == "MiddleButton" then
			-- mod:Debug()
		else
			HudMap:Toggle()
		end			
	end
end

local function groupIter(state, index)
	if index < 0 then return end
	local raid, party = GetNumRaidMembers(), GetNumPartyMembers()
	local prefix = raid > 0 and "raid" or "party"
	local unit = prefix .. index
	if UnitExists(unit) then
		return index + 1, unit
	elseif raid == 0 then
		return -1, "player"
	end
end

local function group()
	return groupIter, nil, 1
end

mod.group = group

local onShow = function(self)
	self.rotSettings = GetCVar("rotateMinimap")
	SetCVar("rotateMinimap", "1")
	if db.useGatherMate and GatherMate then
		GatherMate:GetModule("Display"):ReparentMinimapPins(HudMapStandaloneCluster)
		GatherMate:GetModule("Display"):ChangedVars(nil, "ROTATE_MINIMAP", "1")
	end
	
	if db.useQuestHelper and QuestHelper and QuestHelper.SetMinimapObject then
		QuestHelper:SetMinimapObject(HudMapStandaloneCluster)
	end
	
	if db.useRoutes and Routes and Routes.ReparentMinimap then
		Routes:ReparentMinimap(HudMapStandaloneCluster)
		Routes:CVAR_UPDATE(nil, "ROTATE_MINIMAP", "1")
	end
	
	if TomTom and TomTom.ReparentMinimap then
		TomTom:ReparentMinimap(HudMapStandaloneCluster)
		local Astrolabe = DongleStub("Astrolabe-0.4")
		Astrolabe.processingFrame:SetParent(HudMapStandaloneCluster)
	end
	
	if _G.GetMinimapShape and not mod:IsHooked("GetMinimapShape") then
		mod:Hook("GetMinimapShape")
	end
	
	MinimapCluster:Hide()
end

local onHide = function(self, force)
	SetCVar("rotateMinimap", self.rotSettings)
	if (db.useGatherMate or force) and GatherMate then
		GatherMate:GetModule("Display"):ReparentMinimapPins(Minimap)
		GatherMate:GetModule("Display"):ChangedVars(nil, "ROTATE_MINIMAP", self.rotSettings)
	end
	
	if db.useQuestHelper and QuestHelper and QuestHelper.SetMinimapObject then
		QuestHelper:SetMinimapObject(Minimap)
	end
	
	if (db.useRoutes or force) and Routes and Routes.ReparentMinimap then
		Routes:ReparentMinimap(Minimap)
		Routes:CVAR_UPDATE(nil, "ROTATE_MINIMAP", self.rotSettings)
	end
	
	if TomTom and TomTom.ReparentMinimap then
		TomTom:ReparentMinimap(Minimap)
		local Astrolabe = DongleStub("Astrolabe-0.4")
		Astrolabe.processingFrame:SetParent(Minimap)		
	end
	
	if _G.GetMinimapShape and mod:IsHooked("GetMinimapShape") then
		mod:Unhook("GetMinimapShape")
	end
	
	-- updateFrame:SetScript("OnUpdate", nil)
	MinimapCluster:Show()
end

local optionFrames = {}
local ACD3 = LibStub("AceConfigDialog-3.0")
local coloredTextures = {}
local gatherCircle, gatherLine
local indicators = {"N", "NE", "E", "SE", "S", "SW", "W", "NW"}
local new, free
local pointCache, edgeCache = {}, {}
local activePointList, activeEdgeList = {}, {}

local zoomSpeed = 55
local zoomScale, targetZoomScale = 45, 40
local zoomMin, zoomMax = 15, 100

do
	local fine, coarse, veryCoarse = 1 / 60, 1, 5
	local fineTotal, coarseTotal, veryCoarseTotal = 0, 0, 0
	
	local function computeNewScale()
		local px, py = mod:GetUnitPosition("player")
		local maxDistance = 0
		for point, _ in pairs(activePointList) do
			local d = point:Distance(px, py)
			if d > 0 and d < db.interestRadius and d > maxDistance then
				maxDistance = d
			end
		end
		if maxDistance < 15 then maxDistance = 15 end
		return maxDistance
	end
	
	function updateRotations(self, t)
		fineTotal = fineTotal + t
		coarseTotal = coarseTotal + t
		veryCoarseTotal = veryCoarseTotal + t
		
		if fineTotal > fine then
			while fineTotal > fine do fineTotal = fineTotal - fine end
			
			if zoomScale < targetZoomScale then
				local new = zoomScale + (t * zoomSpeed)
				zoomScale = new > targetZoomScale and targetZoomScale or new 
			elseif zoomScale > targetZoomScale then
				local new = zoomScale - (t * zoomSpeed)
				zoomScale = new < targetZoomScale and targetZoomScale or new 
			end
		
			mod:Update()
			if db.useAdaptiveZoom then
				targetZoomScale = computeNewScale()
			end
		end
		if coarseTotal > coarse then
			while coarseTotal > coarse do coarseTotal = coarseTotal - coarse end
			mod:UpdateZoneData()
		end
		-- if veryCoarseTotal > veryCoarse then
			-- while veryCoarseTotal > veryCoarse do veryCoarseTotal = veryCoarseTotal - veryCoarse end
		-- end
	end
end


function mod:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("HudMapDB", defaults)
	db = self.db.profile
	
	AceConfig:RegisterOptionsTable("HudMap", options, {"hudmap"})
	-- optFrame = AceConfigDialog:AddToBlizOptions("HudMap", "HudMap")
	self:RegisterModuleOptions("General", coreOptions, L["General"])
	self:RegisterModuleOptions("Profiles", LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db), L["Profiles"])
	
	self.canvas = CreateFrame("Frame", "HudMapCanvas", UIParent)
	self.canvas:SetWidth(UIParent:GetWidth())
	self.canvas:SetHeight(UIParent:GetHeight())
	self.canvas:SetPoint("CENTER")
	
	HudMapMinimap:SetPoint("CENTER", UIParent, "CENTER")
	HudMapStandaloneCluster:SetFrameStrata("BACKGROUND")
	-- HudMapStandaloneCluster:SetAlpha(db.alpha)
	HudMapMinimap:SetAlpha(0)
	HudMapMinimap:EnableMouse(false)
	
	setmetatable(HudMapStandaloneCluster, { __index = HudMapMinimap })
	
	HudMapStandaloneCluster:Hide()	
	HudMapStandaloneCluster:SetScript("OnShow", onShow)
	HudMapStandaloneCluster:SetScript("OnHide", onHide)	
	self:HookScript(Minimap, "OnShow", "Minimap_OnShow")
	self:HookScript(MinimapCluster, "OnShow", "Minimap_OnShow")
	
	-- FIXME
	-- parent:RegisterModuleOptions(modName, options, modName)
	-- self:SetScales()
	
	HudMapStandaloneCluster._GetScale = HudMapStandaloneCluster.GetScale
	HudMapStandaloneCluster.GetScale = function()
		return 1
	end
end

function mod:OnEnable()
	db = self.db.profile
	self:RegisterEvent("PLAYER_ENTERING_WORLD",	"UpdateZoneData")
	self:RegisterEvent("ZONE_CHANGED", 					"UpdateZoneData")
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA",	"UpdateZoneData")
	self:RegisterEvent("ZONE_CHANGED_INDOORS", 	"UpdateZoneData")	
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	updateFrame:SetScript("OnUpdate", updateRotations)
	self.canvas:SetAlpha(db.alpha)
	self:UpdateCanvasPosition()
	
	targetZoomScale = db.scale	
	mod.pixelsPerYard = HudMapMinimap:GetWidth() / self:GetMinimapSize()	
	self:UpdateZoneData()
	self:Toggle(db.enabled)
	self:SetZoom()
	-- /run x,y = HudMap:GetUnitPosition("player"); HudMap:PlaceRangeMarker("highlight", x+10, y, 10); HudMap:PlaceRangeMarker("highlight", x+20, y, 10); HudMap:PlaceRangeMarker("highlight", x+20, y+10, 10)
end

do
	local serial = 0
	local lastDot
	local debugging = false
	function mod:ReplaceMarker()
		if not debugging then return end
		serial = (serial + 1) % 8
		local x,y = HudMap:GetUnitPosition("player")
		local mx = math.random(94) - 47
		local my = math.random(94) - 47
		local r, g, b, a = math.random(), math.random(), math.random(), (math.random(128) + 128) / 255
		local lifetime = math.random() * 8
		local radius = math.random() * 8
		local dot = HudMap:PlaceRangeMarker("highlight", x+mx, y+my, radius, lifetime, r, g, b, a):Appear():Rotate(360, lifetime)		
		local edge = HudMap:AddEdge(r, g, b, a, lifetime, "player", nil, nil, nil, x+mx, y+my)
		dot:AttachEdge(edge)
		if lastDot and not lastDot.freed then
			edge = HudMap:AddEdge(r, g, b, a, lifetime, nil, nil, lastDot.stickX, lastDot.stickY, x+mx, y+my)
			dot:AttachEdge(edge)
			lastDot:AttachEdge(edge)
		end
		dot.RegisterCallback(self, "Free", "ReplaceMarker")
		lastDot = dot
	end

	function mod:Debug(num)
		debugging = not debugging
		if not debugging then return end
		for i = 1, (num or 10) do
			self:ReplaceMarker()
		end
	end
end

function mod:RegisterModuleOptions(name, optionTbl, displayName)
	options.args[name] = (type(optionTbl) == "function") and optionTbl() or optionTbl
	if not optionFrames.default then
		optionFrames.default = ACD3:AddToBlizOptions("HudMap", nil, nil, name)
	else
		optionFrames[name] = ACD3:AddToBlizOptions("HudMap", displayName, "HudMap", name)
	end
end

function mod:UpdateCanvasPosition()
	self.canvas:ClearAllPoints()
	configFrame:ClearAllPoints()
	if db.canvasX and db.canvasY then
		configFrame:SetPoint("CENTER", UIParent, "CENTER", db.canvasX, db.canvasY)
		self.canvas:SetPoint("CENTER", UIParent, "CENTER", db.canvasX, db.canvasY)
	else
		configFrame:SetPoint("CENTER", UIParent, "CENTER")
		self.canvas:SetPoint("CENTER", UIParent, "CENTER")
	end
end

function mod:Minimap_OnShow()
	if HudMapStandaloneCluster:IsVisible() then
		HudMapStandaloneCluster:Hide()
	end
end

function mod:COMBAT_LOG_EVENT_UNFILTERED(ev, timestamp, event, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, spellID, ...)
	if event == "UNIT_DIED" then
		for k, v in pairs(followedUnits) do
			if sourceName and k and UnitIsUnit(sourceName, k) then
				v:Free()
			end
		end
	end
end

function mod:Toggle(flag)
	if flag == nil then
		flag = not self.canvas:IsVisible()
	end
	db.enabled = flag
	if flag then
		self.canvas:Show()
	else
		self.canvas:Hide()
	end
end

function mod.SetArea()
	mod:UpdateCanvasPosition()
	local size = db.maxSizeSet and (db.maxSize * 2) or 0.96 * UIParent:GetHeight()
	configFrame.lastSize = size
	configFrame:SetSize(size, size)
	configFrame:Show()
	local x,y = HudMap:GetUnitPosition("player")
	local a, b, c, d
	a = HudMap:PlaceRangeMarker("highlight", x+20, y, 8, nil, 1, 0, 0, 0.5):Appear():Pulse(1.2, 0.5)
	b = HudMap:PlaceRangeMarker("highlight", x-20, y, 8, nil, 0, 1, 0, 0.5):Appear():Pulse(1.2, 1)
	c = HudMap:PlaceRangeMarker("highlight", x, y+20, 8, nil, 0, 0, 1, 0.5):Appear():Pulse(1.2, 1.5)
	d = HudMap:PlaceRangeMarker("highlight", x, y-20, 8, nil, 1, 0, 1, 0.5):Appear():Pulse(1.2, 2)
	tinsert(configFrame.testDots, a)
	tinsert(configFrame.testDots, b)
	tinsert(configFrame.testDots, c)
	tinsert(configFrame.testDots, d)
end

function mod:GetMinimapShape()
	return "ROUND"
end

function mod:SetScales()
	HudMapMinimap:SetZoom(0)
	HudMapMinimap:ClearAllPoints()
	HudMapMinimap:SetPoint("CENTER", UIParent, "CENTER")
	
	HudMapStandaloneCluster:ClearAllPoints()
	HudMapStandaloneCluster:SetPoint("CENTER")
	
	local size = UIParent:GetHeight() / db.scale * 0.9
	HudMapMinimap:SetWidth(size)
	HudMapMinimap:SetHeight(size)
	HudMapStandaloneCluster:SetHeight(size)
	HudMapStandaloneCluster:SetWidth(size)
	
	HudMapStandaloneCluster:SetScale(db.scale)
end

-----------------------------------
--- Points
-----------------------------------

mod.textures = textureLookup

local texCoordLookup = {
	party = {0.525, 0.6, 0.04, 0.2},
	tank = {0.5, 0.75, 0, 1},
	dps = {0.75, 1, 0, 1},
	healer = {0.25, 0.5, 0, 1}
}

local animations = {
	onLoad = function(self)
		self.regionParent = self:GetRegionParent()
	end,
	scale = function(self)
		local p = self:GetProgress()
		local progress = self:GetParent():GetLoopState() == "REVERSE" and (1 - p) or p
		
		if progress < 0 then progress = 0
		elseif progress > 1 then progress = 1
		end
		
		local scale = 1 + ((self.pulseTarget - 1) * progress)
		self.regionParent:SetScale(scale)
	end, 
	alpha = function(self)
		self.regionParent:SetAlpha(self:GetProgress())
	end,
	scaleIn = function(self)
		local scale = 1 + ((1 - self:GetProgress()) * 0.5)
		self.regionParent:SetScale(scale)
	end
}

local function DrawRouteLineCustom(T, C, sx, sy, ex, ey, w, relPoint)
   if (not relPoint) then relPoint = "BOTTOMLEFT"; end

   -- Determine dimensions and center point of line
   local dx,dy = ex - sx, ey - sy;
   local cx,cy = (sx + ex) / 2, (sy + ey) / 2;

   -- Normalize direction if necessary
	 local reverse = dx < 0 and -1 or 1
   if (dx < 0) then
      dx,dy = -dx,-dy;
   end

   -- Calculate actual length of line
   local l = math_sqrt((dx * dx) + (dy * dy));

   -- Quick escape if it's zero length
   if (l == 0) then
      T:SetTexCoord(0,0,0,0,0,0,0,0);
      T:SetPoint("BOTTOMLEFT", C, relPoint, cx,cy);
      T:SetPoint("TOPRIGHT",   C, relPoint, cx,cy);
      return;
   end

   -- Sin and Cosine of rotation, and combination (for later)
   local s,c = -dy / l, dx / l;
   local sc = s * c;

   -- Calculate bounding box size and texture coordinates
   local Bwid, Bhgt, BLx, BLy, TLx, TLy, TRx, TRy, BRx, BRy;
   if (dy >= 0) then
      Bwid = ((l * c) - (w * s)) * TAXIROUTE_LINEFACTOR_2;
      Bhgt = ((w * c) - (l * s)) * TAXIROUTE_LINEFACTOR_2;
      BLx, BLy, BRy = (w / l) * sc, s * s, (l / w) * sc;
      BRx, TLx, TLy, TRx = 1 - BLy, BLy, 1 - BRy, 1 - BLx; 
      TRy = BRx;
   else
      Bwid = ((l * c) + (w * s)) * TAXIROUTE_LINEFACTOR_2;
      Bhgt = ((w * c) + (l * s)) * TAXIROUTE_LINEFACTOR_2;
      BLx, BLy, BRx = s * s, -(l / w) * sc, 1 + (w / l) * sc;
      BRy, TLx, TLy, TRy = BLx, 1 - BRx, 1 - BLx, 1 - BLy;
      TRx = TLy;
   end
	 Bwid = Bwid * reverse
	 Bhgt = Bhgt * reverse

   -- Set texture coordinates and anchors
   T:ClearAllPoints();
   T:SetTexCoord(TLx, TLy, BLx, BLy, TRx, TRy, BRx, BRy);
   T:SetPoint("BOTTOMLEFT", C, relPoint, cx - Bwid, cy - Bhgt);
   T:SetPoint("TOPRIGHT",   C, relPoint, cx + Bwid, cy + Bhgt);
end

local edgeSerial = 0
Edge = {
	Free = function(self)
		if self.freed then return nil end
		self.freed = true
		self.callbacks:Fire("Free", self)
		self:UnregisterAllCallbacks(self)
		for point, _ in pairs(self.points) do
			point:DetachEdge(self)
		end		
		wipe(self.points)
		self.srcPlayer, self.dstPlayer, self.sx, self.sy, self.dx, self.dy = nil, nil, nil, nil, nil, nil
		activeEdgeList[self] = nil
		self.fadeOutGroup:Play()
		tinsert(edgeCache, self)
		return nil
	end,
	New = function(self, r, g, b, a, srcPlayer, dstPlayer, sx, sy, dx, dy, lifetime)
		local t = tremove(edgeCache)
		if not t then
			edgeSerial = edgeSerial + 1
			t = setmetatable({}, {__index = self})
			t.points = {}
			t.serial = "Edge " .. edgeSerial
			t.callbacks = CallbackHandler:New(t)
			t.frame = CreateFrame("Frame", nil, mod.canvas)
			t.frame:SetFrameStrata("LOW")
			t.texture = t.frame:CreateTexture()
			t.texture:SetAllPoints()
			-- local line = "Interface\\TaxiFrame\\UI-Taxi-Line"
			local line = "Interface\\AddOns\\HudMap\\assets\\line"
			t.texture:SetTexture(line)
			
			t.fadeOutGroup = t.frame:CreateAnimationGroup()
			t.fadeOut = t.fadeOutGroup:CreateAnimation("alpha")
			t.fadeOut:SetChange(-1)
			t.fadeOut:SetDuration(0.25)
			t.fadeOut:SetScript("OnFinished", function()
				t.frame:Hide()
			end)
		end
		t.lifetime = type(lifetime) == "number" and GetTime() + lifetime or nil
		t.fadeOutGroup:Stop()
		t.frame:Show()
		t.freed = false
		t.r = r or 1
		t.g = g or 1
		t.b = b or 1
		t.a = a or 1
		t.texture:SetVertexColor(r, g, b, a)
		t.srcPlayer, t.dstPlayer = srcPlayer, dstPlayer
		t.sx, t.sy, t.dx, t.dy = sx, sy, dx, dy
		activeEdgeList[t] = true
		return t
	end,
	AttachPoint = function(self, point)
		self.points[point] = true
	end,
	DetachPoint = function(self, point)
		self.points[point] = nil
	end,	
	TrackFrom = function(self, src_or_x, y)
		if type(src_or_x) == "string" then
			self.srcPlayer = src_or_x
		elseif src_or_x and y then
			self.srcPlayer = nil
			self.sx = src_or_x
			self.sy = y
		end
		return self	end,
	TrackTo = function(self, dst_or_x, y)
		if type(dst_or_x) == "string" then
			self.dstPlayer = dst_or_x
		elseif dst_or_x and y then
			self.dstPlayer = nil
			self.dx = dst_or_x
			self.dy = y
		end
		return self
	end,
	UpdateAll = function(self)
		if(self ~= Edge) then return end
		for t, _ in pairs(activeEdgeList) do
			t:Update()
			t.frame:Show()
		end	
	end,
	Update = function(self)
		if self.lifetime and GetTime() > self.lifetime then
			self:Free()
			return
		end
		local sx, sy, dx, dy		
		if self.srcPlayer then
			sx, sy = mod:GetUnitPosition(self.srcPlayer)
		elseif self.sx and self.sy then
			sx, sy = self.sx, self.sy
		end
		
		if self.dstPlayer then
			dx, dy = mod:GetUnitPosition(self.dstPlayer)
		elseif self.dx and self.dy then
			dx, dy = self.dx, self.dy
		end
		if sx and sy and dx and dy then
			sx, sy = mod:LocationToMinimapOffset(sx, sy)
			dx, dy = mod:LocationToMinimapOffset(dx, dy)
			local ox = mod.canvas:GetWidth() / 2
			local oy = mod.canvas:GetHeight() / 2
			sx = sx + ox
			dx = dx + ox
			sy = sy + oy
			dy = dy + oy
			local odx, ody = dx, dy
			local ax = dx - sx
			local ay = dy - sy
			local hyp = math_pow((ax*ax) + (ay*ay), 0.5)
			local angle = math_atan2(ay, ax)
			local offsetX = (math_cos(angle) * (hyp - 5))
			local offsetY = (math_sin(angle) * (hyp - 5))
			dx = sx + offsetX
			dy = sy + offsetY
			sx = odx - offsetX
			sY = ody - offsetY
			if hyp > 15 then
				self.texture:Show()
				DrawRouteLineCustom(self.texture, mod.canvas, sx, sy, dx, dy, 64);
			else
				self.texture:Hide()
			end
		end
	end
}

function mod:AddEdge(r, g, b, a, lifetime, srcPlayer, dstPlayer, sx, sy, dx, dy)
	return Edge:New(r, g, b, a, srcPlayer, dstPlayer, sx, sy, dx, dy, lifetime)
end

local circleSerial = 0
Point = {
	Free = function(self, noAnimate)
		if self.freed then
			return nil
		end
		self.freed = true
		if self.follow then
			followedUnits[self.follow] = nil
		end
		for edge, _ in pairs(self.edges) do
			edge:Free()
		end
		wipe(self.edges)
		
		if noAnimate then
			self.frame:Hide()
			self.pulseAnimations:Stop()
			self.repeatAnimations:Stop()
		else
			self.fadeOutGroup:Play()
		end
		self.callbacks:Fire("Free", self)
		self:UnregisterAllCallbacks(self)

		self.stick = false
		self.stickX = nil
		self.stickY = nil
		self.follow = nil
		self.lifetime = nil
		self.lastPPY = nil
		self.lastRadius = nil
		
		activePointList[self] = nil
		tinsert(pointCache, self)
		return nil
	end,
	
	AttachEdge = function(self, edge)
		self.edges[edge] = true
		edge:AttachPoint(self)
	end,
	
	DetachEdge = function(self, edge)
		self.edges[edge] = nil
		edge:DetachPoint(self)
	end,	
	
	Stick = function(self, zone, x, y)
		self.stick = true
		self.follow = nil
		self.stickX = x
		self.stickY = y
		return self
	end,
	
	Follow = function(self, unit)
		self.stick = false
		self.follow = unit
		followedUnits[unit] = self
		return self
	end,
	
	Update = function(self)
		if self.zone and mod.currentZone ~= self.zone then self:Free(); return end
		if not self.lifetime or self.lifetime > 0 and self.lifetime < GetTime() then self:Free(); return end
		local x, y
		if self.stick then
			x, y = mod:LocationToMinimapOffset(self.stickX, self.stickY)
		elseif self.follow then
			local ux, uy = mod:GetUnitPosition(self.follow)
			if not ux or not uy or (ux == 0 and uy == 0) then
				self:Free()
				return
			end
			x, y = mod:LocationToMinimapOffset(ux, uy)
		end
		
		if not self.frame:IsVisible() then self.frame:Show() end
		
		local needUpdate = false
		if self.follow == "player" then
			needUpdate = not self.placed
		else
			needUpdate = self.lastX ~= x or self.lastY ~= y
		end
		
		if self.radius then
			if self.lastPPY ~= mod.pixelsPerYard or self.lastRadius ~= self.radius then
				self.lastPPY = mod.pixelsPerYard
				self.lastRadius = self.radius
				local pixels = mod:RangeToPixels(self.radius * 2)
				self.frame:SetWidth(pixels)
				self.frame:SetHeight(pixels)
			end
		end
		
		if needUpdate then
			self.frame:ClearAllPoints()
			self.frame:SetPoint("CENTER", self.frame:GetParent(), "CENTER", x, y)
		end
		self.placed = true
		self.lastX = x
		self.lastY = y
		-- self.callbacks:Fire("Update", self, cx, cy, self.radius)
		if self.shouldUpdateRange then
			self:UpdateAlerts()
		end
	end,
	
	UpdateAll = function(self)
		if(self ~= Point) then return end
		for t, _ in pairs(activePointList) do
			t:Update()
			t.frame:Show()
		end
	end,
	
	Pulse = function(self, size, speed)
		self.pulse:SetDuration(speed)
		self.pulse:SetScale(size, size)
		self.pulseIn:SetDuration(speed)
		self.pulseIn:SetScale(1 / size, 1 / size)
		self.pulseAnimations:Play()
		return self
	end,

	Rotate = function(self, amount, speed)
		self.rotate:SetDuration(speed)
		self.rotate:SetDegrees(-amount)
		self.repeatAnimations:Play()		
		return self
	end,
	
	Appear = function(self)
		self.fadeInGroup:Play()
		return self
	end,
	
	SetTexCoords = function(self, a, b, c, d)
		self.texture:SetTexCoord(a,b,c,d)
		return self
	end,
	
	Alert = function(self, bool)
		local r, g, b, a
		r = bool and self.alert.r or self.normal.r or 1
		g = bool and self.alert.g or self.normal.g or 1
		b = bool and self.alert.b or self.normal.b or 1
		a = bool and self.alert.a or self.normal.a or 1
		self.texture:SetVertexColor(r, g, b, a)
		return self
	end,
	
	RegisterForAlerts = function(self, bool)
		if bool == nil then bool = true end
		self.shouldUpdateRange = bool
		return self
	end,
	
	Distance = function(self, x2, y2)
		local x, y, ux, uy
		if self.stick then
			x, y = self.stickX, self.stickY
		elseif self.follow then
			x, y = mod:GetUnitPosition(self.follow)
			if not x or not y or (x == 0 and y == 0) then
				return -1
			end
		end
		local e = x2-x
		local f = y2-y
		return math_sqrt((e*e)+(f*f)) + (self.radius or 0), x, y
	end,
	
	UpdateAlerts = function(self)
		if not self.radius then return end
		local x, y
		if self.stick then
			x, y = self.stickX, self.stickY
		else
			x, y = mod:GetUnitPosition(self.follow)
		end
		
		local alert = false
		if self.shouldUpdateRange == "all" then
			for index, unit in group() do
				alert = mod:DistanceToPoint(unit, x, y) < self.radius
				if alert then break end
			end
		else
			alert = mod:DistanceToPoint("player", x, y) < self.radius
		end
		self:Alert(alert)	
	end,
	
	SetColor = function(self, r, g, b, a)
		self.normal.r = r or 1
		self.normal.g = g or 1
		self.normal.b = b or 1
		self.normal.a = a or 0.5
		self:Alert(false)
		return self
	end,
	
	SetAlertColor = function(self, r, g, b, a)
		self.alert.r = r or 1
		self.alert.g = g or 0
		self.alert.b = b or 0
		self.alert.a = a or 0.5
		return self
	end,
	
	SetTexture = function(self, texfile, blend)
		local tex = self.texture
		texfile = texfile or "glow"
		tex:SetTexture(textureLookup[texfile] or texfile or [[Interface\GLUES\MODELS\UI_Tauren\gradientCircle]])
		if texCoordLookup[texfile] then
			tex:SetTexCoord(unpack(texCoordLookup[texfile]))
		else
			tex:SetTexCoord(0, 1, 0, 1)
		end	
		blend = blend or texBlending[texfile] or "BLEND"
		tex:SetBlendMode(blend)
		return self
	end,
	
	SetSize = function(self, size)
		self.radius, self.fixedSize = nil, nil
		if type(size) == "string" then
			self.fixedSize = size:match("(%d+)px")
			if self.fixedSize then self.fixedSize = tonumber(self.fixedSize) end
			
			self.radius = size:match("(%d+)yd")
			if self.radius then self.radius = tonumber(self.radius) + 2 end
		else
			self.radius = size + 2
		end
		local pixels = self.fixedSize and self.fixedSize or mod:RangeToPixels(self.radius or 1)
		self.frame:SetWidth(pixels)
		self.frame:SetHeight(pixels)
		
		return self
	end,
	
	New = function(self, zone, x, y, follow, lifetime, texfile, size, blend, r, g, b, a)
		local t = tremove(pointCache)
		if not t then
			circleSerial = circleSerial + 1
			t = setmetatable({}, {__index = self})
			t.serial = "Circle " .. circleSerial
			t.callbacks = CallbackHandler:New(t)
			t.frame = CreateFrame("Frame", nil, mod.canvas)
			t.frame:SetFrameStrata("MEDIUM")
			t.edges = {}
			-- t.frame:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background", 
                                            -- edgeFile = "Interface/Tooltips/UI-Tooltip-Border", 
                                            -- tile = true, tileSize = 16, edgeSize = 16, 
                                            -- insets = { left = 0, right = 0, top = 0, bottom = 0 }})
			t.texture = t.frame:CreateTexture()
			t.texture:SetAllPoints()
			t.repeatAnimations = t.frame:CreateAnimationGroup()
			t.repeatAnimations:SetLooping("REPEAT")
			
			t.pulseAnimations = t.frame:CreateAnimationGroup()
			-- t.pulseAnimations:SetLooping("BOUNCE")
			t.pulseAnimations:SetScript("OnFinished", function(self)
				self:Play()
			end)
			
			t.pulse = t.pulseAnimations:CreateAnimation("scale")
			t.pulse:SetOrder(1)
			t.pulseIn = t.pulseAnimations:CreateAnimation("scale")
			t.pulseIn:SetOrder(2)
			t.pulse:SetScript("OnPlay", animations.onLoad)
			
			t.rotate = t.repeatAnimations:CreateAnimation("rotation")
			
			t.normal, t.alert = {}, {}
			
			do
				t.fadeInGroup = t.frame:CreateAnimationGroup()
				
				local scaleOut = t.fadeInGroup:CreateAnimation("scale")
				scaleOut:SetDuration(0)
				scaleOut:SetScale(1.5, 1.5)
				scaleOut:SetOrder(1)
				
				t.fadeIn = t.fadeInGroup:CreateAnimation()
				t.fadeIn:SetDuration(0.35)
				t.fadeIn:SetScript("OnPlay", animations.onLoad)
				t.fadeIn:SetScript("OnUpdate", animations.alpha)
				t.fadeIn:SetOrder(2)

				local scaleIn = t.fadeInGroup:CreateAnimation("scale")
				scaleIn:SetDuration(0.35)
				scaleIn:SetScale(1 / 1.5, 1 / 1.5)
				scaleIn:SetOrder(2)
			end
			
			t.fadeOutGroup = t.frame:CreateAnimationGroup()
			t.fadeOut = t.fadeOutGroup:CreateAnimation("alpha")
			t.fadeOut:SetChange(-1)
			t.fadeOut:SetDuration(0.25)
			t.fadeOut:SetScript("OnFinished", function()
				t.frame:Hide()
				t.pulseAnimations:Stop()
				t.repeatAnimations:Stop()
			end)
			
			-- t.pulse:SetSmoothing("OUT")
			-- t.rotate:SetSmoothing("OUT")
		end
		if t.freed == false then
			error("ERROR: Attempted to reallocate a freed frame.")
		end
		t.freed = false
		t.fadeOutGroup:Stop()
		t.frame:SetAlpha(1)
		t.placed = false
		t.frame:Show()		
		
		t:SetSize(size or 20)
		t:SetTexture(texfile, blend)
		
		t:SetColor(r, g, b, a)
		t:SetAlertColor(r, g, b, a)
		t:Alert(false)
		
		t.shouldUpdateRange = false
		
		if x and y then
			t:Stick(zone, x, y)
		elseif follow then
			t:Follow(follow)
		end
		t.lifetime = lifetime and (GetTime() + lifetime) or -1
		t.zone = zone
		activePointList[t] = true
		return t
	end,
}

function mod:PlaceRangeMarker(texture, x, y, radius, duration, r, g, b, a, blend)
	-- print("Placing", texture, "at", x, ", ", y, "r", radius)
	return Point:New(self.currentZone, x, y, nil, duration, texture, radius, blend, r, g, b, a)	
end

function mod:PlaceRangeMarkerOnPartyMember(texture, person, radius, duration, r, g, b, a, blend)
	-- print("Placing", texture, "on", person, "r", radius)
	return Point:New(nil, nil, nil, person, duration, texture, radius, blend, r, g, b, a)
end

local ABS, POW = math.abs, math.pow
function mod:DistanceToPoint(unit, x, y)
	local x1, y1 = self:GetUnitPosition(unit)
	local x2, y2 = x, y
	local dx = x2 - x1
	local dy = y2 - y1
	return ABS(POW((dx*dx)+(dy*dy), 0.5))
end

function mod:UnitDistance(unitA, unitB)
	local x1, y1 = self:GetUnitPosition(unitA)
	local x2, y2 = self:GetUnitPosition(unitB)
	local dx = x2 - x1
	local dy = y2 - y1
	-- 
	return ABS(POW((dx*dx)+(dy*dy), 0.5))
end

function mod:GetUnitPosition(unit)
	if not unit then return nil, nil end
	local x, y = GetPlayerMapPosition(unit)
	return self:CoordsToPosition(x, y)
end

function mod:CoordsToPosition(x, y)
	if not x or not y or (x == 0 and y == 0) then return x, y end
	if not self.zoneScale then
		return x * 1500, (1 - y) * 1000
	end
	return x * self.zoneScale[1], (1 - y) * self.zoneScale[2]
end

function mod:UpdateZoneData()
	SetMapToCurrentZone()
	local area, level, key
	area = GetMapInfo()
	level = GetCurrentMapDungeonLevel()
	
	-- Thanks Cyprias!
	if area == "Ulduar" or area == "CoTStratholme" then
		level = level - 1
	end
	key = level > 0 and (area .. level) or area
	self.currentZone = zoneOverrides[GetSubZoneText()] or key	
	self.zoneScale = zoneScalingData[self.currentZone]
end

do
	local a, b
	function mod:Measure(restart)
		local a2, b2 = self:GetUnitPosition("player")
		if restart or not a then
			a, b = a2, b2
		else
			local c, d = (a2 - a), (b2 - b)
			print(math.sqrt((c*c)+(d*d)))
			a, b = nil, nil
		end
	end
end

function mod:SetZoom(zoom, zoomChange)
	if zoom then
		targetZoomScale = zoom
	elseif zoomChange then
		targetZoomScale = targetZoomScale + zoomChange
	else
		targetZoomScale = db.zoomLevel
	end
	db.zoomLevel = targetZoomScale
	if targetZoomScale < 20 then
		targetZoomScale = 20
	elseif targetZoomScale > 200 then
		targetZoomScale = 200
	end
end

function mod:Update()
	Point:UpdateAll()
	Edge:UpdateAll()
end

function mod:GetMinimapSize()
	return zoomScale -- math.pow(zoomScale, 2)
	-- return minimapSize[indoors and "indoor" or "outdoor"][HudMapMinimap:GetZoom()]
end

do
	function mod:LocationToMinimapOffset(x, y)
		mod.pixelsPerYard = db.maxSize / zoomScale
		local px, py = self:GetUnitPosition("player")
		local dx = (px - x) * mod.pixelsPerYard
		local dy = (py - y) * mod.pixelsPerYard
		
		-- Now adjust for rotation
		local bearing = GetPlayerFacing()
		local angle = math_atan2(dx, dy)
		local hyp = math.abs(math_sqrt((dx * dx) + (dy * dy)))
		local x, y = math_sin(angle + bearing), math_cos(angle + bearing)
		return -x * hyp, -y * hyp
	end

	function mod:RangeToPixels(range)
		local px, py = self:GetUnitPosition("player")
		local ay, ay = mod:LocationToMinimapOffset(px, py)
		local by, by = mod:LocationToMinimapOffset(px, py+10)
		local comp = by - ay
		mod.pixelsPerYard = db.maxSize / zoomScale
		return mod.pixelsPerYard * range
	end
end

------------------------------------
-- Data
------------------------------------

minimapSize = {
	indoor = {
		[0] = 290,
		[1] = 230,
		[2] = 175,
		[3] = 119,
		[4] = 79,
		[5] = 49.8,
	},
	outdoor = {
		[0] = 450,
		[1] = 395,
		[2] = 326,
		[3] = 265,
		[4] = 198,
		[5] = 132
	},
}
mod.minimapSize = minimapSize

zoneOverrides = {
	[L["The Frozen Throne"]] = "IcecrownCitadel7"
}

zoneScalingData = setmetatable({
	Arathi = { 3599.99987792969, 2399.99992370605, 1},
	Ogrimmar = { 1402.6044921875, 935.416625976563, 2},
	Undercity = { 959.375030517578, 640.104125976563, 4},
	Barrens = { 10133.3330078125, 6756.24987792969, 5},
	Darnassis = { 1058.33325195313, 705.7294921875, 6},
	AzuremystIsle = { 4070.8330078125, 2714.5830078125, 7},
	UngoroCrater = { 3699.99981689453, 2466.66650390625, 8},
	BurningSteppes = { 2929.16659545898, 1952.08349609375, 9},
	Wetlands = { 4135.41668701172, 2756.25, 10},
	Winterspring = { 7099.99984741211, 4733.33325195313, 11},
	Dustwallow = { 5250.00006103516, 3499.99975585938, 12},
	Darkshore = { 6549.99975585938, 4366.66650390625, 13},
	LochModan = { 2758.33312988281, 1839.5830078125, 14},
	BladesEdgeMountains = { 5424.99975585938, 3616.66638183594, 15},
	Durotar = { 5287.49963378906, 3524.99987792969, 16},
	Silithus = { 3483.333984375, 2322.916015625, 17},
	ShattrathCity = { 1306.25, 870.833374023438, 18},
	Ashenvale = { 5766.66638183594, 3843.74987792969, 19},
	Azeroth = { 40741.181640625, 27149.6875, 20},
	Nagrand = { 5525.0, 3683.33316802979, 21},
	TerokkarForest = { 5399.99975585938, 3600.00006103516, 22},
	EversongWoods = { 4925.0, 3283.3330078125, 23},
	SilvermoonCity = { 1211.45849609375, 806.7705078125, 24},
	Tanaris = { 6899.99952697754, 4600.0, 25},
	Stormwind = { 1737.499958992, 1158.3330078125, 26},
	SwampOfSorrows = { 2293.75, 1529.1669921875, 27},
	EasternPlaguelands = { 4031.25, 2687.49987792969, 28},
	BlastedLands = { 3349.99987792969, 2233.333984375, 29},
	Elwynn = { 3470.83325195313, 2314.5830078125, 30},
	DeadwindPass = { 2499.99993896484, 1666.6669921875, 31},
	DunMorogh = { 4924.99975585938, 3283.33325195313, 32},
	TheExodar = { 1056.7705078125, 704.687744140625, 33},
	Felwood = { 5749.99963378906, 3833.33325195313, 34},
	Silverpine = { 4199.99975585938, 2799.99987792969, 35},
	ThunderBluff = { 1043.74993896484, 695.833312988281, 36},
	Hinterlands = { 3850.0, 2566.66662597656, 37},
	StonetalonMountains = { 4883.33312988281, 3256.24981689453, 38},
	Mulgore = { 5137.49987792969, 3424.99984741211, 39},
	Hellfire = { 5164.5830078125, 3443.74987792969, 40},
	Ironforge = { 790.625061035156, 527.6044921875, 41},
	ThousandNeedles = { 4399.99969482422, 2933.3330078125, 42},
	Stranglethorn = { 6381.24975585938, 4254.166015625, 43},
	Badlands = { 2487.5, 1658.33349609375, 44},
	Teldrassil = { 5091.66650390625, 3393.75, 45},
	Moonglade = { 2308.33325195313, 1539.5830078125, 46},
	ShadowmoonValley = { 5500.0, 3666.66638183594, 47},
	Tirisfal = { 4518.74987792969, 3012.49981689453, 48},
	Aszhara = { 5070.83276367188, 3381.24987792969, 49},
	Redridge = { 2170.83325195313, 1447.916015625, 50},
	BloodmystIsle = { 3262.4990234375, 2174.99993896484, 51},
	WesternPlaguelands = { 4299.99990844727, 2866.66653442383, 52},
	Alterac = { 2799.99993896484, 1866.66665649414, 53},
	Westfall = { 3499.99981689453, 2333.3330078125, 54},
	Duskwood = { 2699.99993896484, 1800.0, 55},
	Netherstorm = { 5574.99967193604, 3716.66674804688, 56},
	Ghostlands = { 3300.0, 2199.99951171875, 57},
	Zangarmarsh = { 5027.08349609375, 3352.08325195313, 58},
	Desolace = { 4495.8330078125, 2997.91656494141, 59},
	Kalimdor = { 36799.810546875, 24533.2001953125, 60},
	SearingGorge = { 2231.24984741211, 1487.49951171875, 61},
	Expansion01 = { 17464.078125, 11642.71875, 62},
	Feralas = { 6949.99975585938, 4633.3330078125, 63},
	Hilsbrad = { 3199.99987792969, 2133.33325195313, 64},
	Sunwell = { 3327.0830078125, 2218.7490234375, 65},
	Northrend = { 17751.3984375, 11834.2650146484, 66},
	BoreanTundra = { 5764.5830078125, 3843.74987792969, 67},
	Dragonblight = { 5608.33312988281, 3739.58337402344, 68},
	GrizzlyHills = { 5249.99987792969, 3499.99987792969, 69},
	HowlingFjord = { 6045.83288574219, 4031.24981689453, 70},
	IcecrownGlacier = { 6270.83331298828, 4181.25, 71},
	SholazarBasin = { 4356.25, 2904.16650390625, 72},
	TheStormPeaks = { 7112.49963378906, 4741.666015625, 73},
	ZulDrak = { 4993.75, 3329.16650390625, 74},
	ScarletEnclave = { 3162.5, 2108.33337402344, 76},
	CrystalsongForest = { 2722.91662597656, 1814.5830078125, 77},
	LakeWintergrasp = { 2974.99987792969, 1983.33325195313, 78},
	StrandoftheAncients = { 1743.74993896484, 1162.49993896484, 79},
	Dalaran = { 0.0, 0.0, 80},
	Naxxramas = { 1856.24975585938, 1237.5, 81},
	Naxxramas1 = { 1093.830078125, 729.219970703125, 82},
	Naxxramas2 = { 1093.830078125, 729.219970703125, 83},
	Naxxramas3 = { 1200.0, 800.0, 84},
	Naxxramas4 = { 1200.330078125, 800.219970703125, 85},
	Naxxramas5 = { 2069.80981445313, 1379.8798828125, 86},
	Naxxramas6 = { 655.93994140625, 437.2900390625, 87},
	TheForgeofSouls = { 11399.9995117188, 7599.99975585938, 88},
	TheForgeofSouls1 = { 1448.09985351563, 965.400390625, 89},
	AlteracValley = { 4237.49987792969, 2824.99987792969, 90},
	WarsongGulch = { 1145.83331298828, 764.583312988281, 91},
	IsleofConquest = { 2650.0, 1766.66658401489, 92},
	TheArgentColiseum = { 2599.99996948242, 1733.33334350586, 93},
	TheArgentColiseum1 = { 369.986186981201, 246.657989501953, 94},
	TheArgentColiseum1 = { 369.986186981201, 246.657989501953, 95},
	TheArgentColiseum2 = { 739.996017456055, 493.330017089844, 96},
	HrothgarsLanding = { 3677.08312988281, 2452.083984375, 97},
	AzjolNerub = { 1072.91664505005, 714.583297729492, 98},
	AzjolNerub1 = { 752.973999023438, 501.983001708984, 99},
	AzjolNerub2 = { 292.973999023438, 195.315979003906, 100},
	AzjolNerub3 = { 367.5, 245.0, 101},
	Ulduar77 = { 3399.99981689453, 2266.66666412354, 102},
	Ulduar771 = { 920.196014404297, 613.466064453125, 103},
	DrakTharonKeep = { 627.083312988281, 418.75, 104},
	DrakTharonKeep1 = { 619.941009521484, 413.293991088867, 105},
	DrakTharonKeep2 = { 619.941009521484, 413.293991088867, 106},
	HallsofReflection = { 12999.9995117188, 8666.66650390625, 107},
	HallsofReflection1 = { 879.02001953125, 586.01953125, 108},
	TheObsidianSanctum = { 1162.49991798401, 775.0, 109},
	HallsofLightning = { 3399.99993896484, 2266.66666412354, 110},
	HallsofLightning1 = { 566.235015869141, 377.489990234375, 111},
	HallsofLightning2 = { 708.237014770508, 472.160034179688, 112},
	IcecrownCitadel = { 12199.9995117188, 8133.3330078125, 113},
	IcecrownCitadel1 = { 1355.47009277344, 903.647033691406, 114},
	IcecrownCitadel2 = { 1067.0, 711.333690643311, 115},
	IcecrownCitadel3 = { 195.469970703125, 130.315002441406, 116},
	IcecrownCitadel4 = { 773.710083007813, 515.810302734375, 117},
	IcecrownCitadel5 = { 1148.73999023438, 765.820068359375, 118},
	IcecrownCitadel6 = { 373.7099609375, 249.1298828125, 119},
	IcecrownCitadel7 = { 293.260009765625, 195.507019042969, 120},
	IcecrownCitadel8 = { 247.929931640625, 165.287994384766, 121},
	VioletHold = { 383.333312988281, 256.25, 122},
	VioletHold1 = { 256.22900390625, 170.820068359375, 123},
	NetherstormArena = { 2270.83319091797, 1514.58337402344, 124},
	CoTStratholme = { 1824.99993896484, 1216.66650390625, 125},
	CoTStratholme1 = { 1125.29998779297, 750.199951171875, 126},
	TheEyeofEternity = { 3399.99981689453, 2266.66666412354, 127},
	TheEyeofEternity1 = { 430.070068359375, 286.713012695313, 128},
	Nexus80 = { 2600.0, 1733.33322143555, 129},
	Nexus801 = { 514.706970214844, 343.138977050781, 130},
	Nexus802 = { 664.706970214844, 443.138977050781, 131},
	Nexus803 = { 514.706970214844, 343.138977050781, 132},
	Nexus804 = { 294.700988769531, 196.463989257813, 133},
	VaultofArchavon = { 2599.99987792969, 1733.33325195313, 134},
	VaultofArchavon1 = { 1398.25500488281, 932.170013427734, 135},
	Ulduar = { 3287.49987792969, 2191.66662597656, 136},
	Ulduar1 = { 669.450988769531, 446.300048828125, 137},
	Ulduar2 = { 1328.46099853516, 885.639892578125, 138},
	Ulduar3 = { 910.5, 607.0, 139},
	Ulduar4 = { 1569.4599609375, 1046.30004882813, 140},
	Ulduar5 = { 619.468994140625, 412.97998046875, 141},
	Dalaran1 = { 830.015014648438, 553.33984375, 142},
	Dalaran2 = { 563.223999023438, 375.48974609375, 143},
	Gundrak = { 1143.74996948242, 762.499877929688, 144},
	Gundrak1 = { 905.033050537109, 603.35009765625, 145},
	TheNexus = { 0.0, 0.0, 146},
	TheNexus1 = { 1101.2809753418, 734.1875, 147},
	PitofSaron = { 1533.33331298828, 1022.91667175293, 148},
	Ahnkahet = { 972.91667175293, 647.916610717773, 149},
	Ahnkahet1 = { 972.41796875, 648.279022216797, 150},
	ArathiBasin = { 1756.24992370605, 1170.83325195313, 151},
	UtgardePinnacle = { 6549.99951171875, 4366.66650390625, 152},
	UtgardePinnacle1 = { 548.936019897461, 365.957015991211, 153},
	UtgardePinnacle2 = { 756.179943084717, 504.119003295898, 154},
	UtgardeKeep = { 0.0, 0.0, 155},
	UtgardeKeep1 = { 734.580993652344, 489.721500396729, 156},
	UtgardeKeep2 = { 481.081008911133, 320.720293045044, 157},
	UtgardeKeep3 = { 736.581008911133, 491.054512023926, 158},
}, {__index = function(t, k)
	if k then
		error("HudMap has no zone data for " .. k .. ". Please report this as a bug.")
		rawset(t, k, false)
	end
	return rawget(t, k)
end })